/*
 *  TOPPERS Software
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 *
 *  Copyright (C) 2005-2008 by TAKAGI Nobuhisa
 * 
 *  L쌠҂́Cȉ(1)`(4)̏𖞂ꍇɌC{\tgEF
 *  Ai{\tgEFAς̂܂ށDȉjgpEE
 *  ρEĔzziȉCpƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *      ܂C{\tgEFÃ[U܂̓Gh[ÛȂ闝
 *      RɊÂCL쌠҂TOPPERSvWFNg
 *      Ɛӂ邱ƁD
 * 
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̎gpړI
 *  ɑ΂K܂߂āCȂۏ؂sȂD܂C{\tgEF
 *  A̗pɂ蒼ړI܂͊ԐړIɐȂ鑹QɊւĂC
 *  ̐ӔC𕉂ȂD
 * 
 */

/*
 *  toppers/s_record.cpp
 */
#include "toppers/s_record.hpp"
#include <istream>
#include <algorithm>
#include <cctype>
#include <cstring>

namespace toppers
{

  /*!
   *  \brief  SR[h̃[h
   *  \param  istr  ̓Xg[
   */
  void s_record::load( std::istream& istr )
  {
    int type = -1;

    while ( istr )
    {
      std::string buf;
      std::getline( istr, buf );
      if ( buf.empty() )
      {
        break;
      }
      record rec = read_record( buf );

      // ܂茵ɂ͏Ȃ
      if ( '1' <= rec.type && rec.type <= '3' )
      {
        bool done = false;
        typedef std::vector< value_type >::iterator iterator;
        for ( iterator iter( data_.begin() ), last( data_.end() ); iter != last; ++iter )
        {
          if ( rec.address == iter->first + iter->second.size() )
          {
            std::copy( rec.data.begin(), rec.data.end(), std::back_inserter( iter->second ) );
            done = true;
          }
        }
        if ( !done )
        {
          data_.push_back( value_type( rec.address, std::vector< unsigned char >( rec.data.begin(), rec.data.end() ) ) );
        }
      }
      type = rec.type;
    }

    cache_ = data_.begin();
  }

  /*!
   *  \brief  wAhX̃oCgf[^擾
   *  \param  address   AhXw
   *  \return address Ŏw肵AhX̃oCgf[^Ԃ
   */
  int s_record::operator[]( size_type address ) const
  {
    typedef std::vector< value_type >::const_iterator const_iterator;
    if ( cache_ != data_.end() )
    {
      const_iterator iter( cache_ );
      if ( iter->first <= address && address < iter->first + iter->second.size() )
      {
        return iter->second.at( address - iter->first );
      }
    }
    for ( const_iterator iter( cache_ ), last( data_.end() ); iter != last; ++iter )
    {
      if ( iter->first <= address && address < iter->first + iter->second.size() )
      {
        cache_ = iter;
        return iter->second.at( address - iter->first );
      }
    }
    for ( const_iterator iter( data_.begin() ), last( cache_ ); iter != last; ++iter )
    {
      if ( iter->first <= address && address < iter->first + iter->second.size() )
      {
        cache_ = iter;
        return iter->second.at( address - iter->first );
      }
    }
    return -1;
  }

  /*!
   *  \brief  SR[h̐lǂݍ
   *  \param  base  ǂݍ݈ʒu̐擪AhX
   *  \param  size  l̃oCg
   *  \param  little_endian gGfBAȂ trueArbOGfBAȂ false w
   *  \return ǂݍ񂾐lԂ
   */
  boost::uintmax_t s_record::get_value( std::size_t base, std::size_t size, bool little_endian ) const
  {
    boost::uintmax_t value = 0;
    if ( little_endian )
    {
      for ( long j = static_cast< long >( size-1 ); j >= 0; j-- )
      {
        int t = ( *this )[ base + j ];
        if ( t < 0 )
        {
          throw data_error();
        }
        value = ( value << 8 ) | ( t & 0xff );
      }
    }
    else
    {
      for ( std::size_t j = 0; j < size; j++ )
      {
        int t = ( *this )[ base + j ];
        if ( t < 0 )
        {
          throw data_error();
        }
        value = ( value << 8 ) | ( t & 0xff );
      }
    }
    return value;
  }

  unsigned long s_record::lower_bound() const
  {
    return data_.front().first;
  }

  unsigned long s_record::upper_bound() const
  {
    return data_.back().first + data_.back().second.size();
  }

  /*!
   *  \brief  SR[ĥPsǂݍ
   *  \param  rec_buf Psobt@
   *  \return ǂݍ݌ʂԂ
   */
  s_record::record const s_record::read_record( std::string const& rec_buf )
  {
    std::string buf( rec_buf );

    // s'\r'܂'\n'cĂꍇ̑΍
    while ( std::isspace( static_cast< unsigned char >( *buf.rbegin() ) ) )
    {
      buf = buf.substr( 0, buf.size()-1 );
    }

    if ( buf.size() < 10 || buf[0] != 'S' )
    {
      throw format_error();
    }
    int ch = static_cast< unsigned char >( buf[1] );
    int address_length = 4;
    std::string::size_type size = buf.size();

    switch ( ch )
    {
    case '1':
    case '9':
      address_length = 4;
      break;
    case '2':
    case '8':
      address_length = 6;
      break;
    case '3':
    case '7':
      address_length = 8;
      break;
    default:
      if ( !std::isdigit( static_cast< unsigned char >( ch ) ) )
      {
        throw format_error();
      }
      break;
    }
    
    record rec;
    rec.type = ch;
    rec.length = xdigit_to_int( buf[2] ) << 4 | xdigit_to_int( buf[3] );
    if ( rec.length * 2u + 4 != buf.size() )
    {
      throw format_error();
    }
    rec.length -= address_length/2 + 1; // AhXƃ`FbNT̒āAf[^ɒ

    rec.address = 0;
    int base = 4;
    for ( int i = 0; i < address_length; i++ )
    {
      rec.address = rec.address << 4 | xdigit_to_int( buf[base+i] );
    }

    base += address_length;
    rec.data.reserve( rec.length );
    for ( int i = 0; i < rec.length; i++ )
    {
      rec.data.push_back( static_cast< unsigned char >( xdigit_to_int( buf[base+i*2] ) << 4 | xdigit_to_int( buf[base+i*2+1] ) ) );
    }

    rec.checksum = xdigit_to_int( buf[size-2] ) << 4 | xdigit_to_int( buf[size-1] );

    // `FbNT͏ȗ

    return rec;
  }

  /*!
   *  \brief  \Zi琔lւ̕ϊ
   *  \param  ch  \Ziij
   *  \return ch ɑΉ鐔l
   */
  int s_record::xdigit_to_int( int ch )
  {
    static char const xdigits[] = "0123456789abcdef";

    ch = std::tolower( static_cast< unsigned char >( ch ) );
    char const* s = std::strchr( xdigits, ch );
    if ( s == 0 )
    {
      return -1;
    }
    return s - xdigits;
  }

}
